from nose.tools import eq_
from pylibmc.client import translate_server_spec
import _pylibmc

def test_translate():
    eq_(translate_server_spec("111.122.133.144"),
        (_pylibmc.server_type_tcp, "111.122.133.144", 11211, 1))

def test_translate():
    eq_(translate_server_spec("111.122.133.144:5555"),
        (_pylibmc.server_type_tcp, "111.122.133.144", 5555, 1))

def test_udp_translate():
    eq_(translate_server_spec("udp:199.299.399.499:5555"),
        (_pylibmc.server_type_udp, "199.299.399.499", 5555, 1))

def test_udp_translate_ipv6():
    eq_(translate_server_spec("udp:[abcd:abcd::1]:5555"),
        (_pylibmc.server_type_udp, "abcd:abcd::1", 5555, 1))

def test_translate_with_weight_string():
    eq_(translate_server_spec("111.122.133.144:5555:2"),
        (_pylibmc.server_type_tcp, "111.122.133.144", 5555, 2))

def test_translate_with_weight_kwd():
    eq_(translate_server_spec("111.122.133.144", weight=2),
        (_pylibmc.server_type_tcp, "111.122.133.144", 11211, 2))

def test_udp_translate_ipv6_with_weight():
    eq_(translate_server_spec("udp:[abcd:abcd::1]:5555:2"),
        (_pylibmc.server_type_udp, "abcd:abcd::1", 5555, 2))




"""Python-level wrapper client"""

import _pylibmc
from .consts import (hashers, distributions, all_behaviors,
                     hashers_rvs, distributions_rvs,
                     all_callbacks, BehaviorDict)

_all_behaviors_set = set(all_behaviors)
_all_behaviors_set.update(set(all_callbacks))

server_type_map = {"tcp":   _pylibmc.server_type_tcp,
                   "udp":   _pylibmc.server_type_udp,
                   "unix":  _pylibmc.server_type_unix}

def _split_spec_type(spec):
    if spec.startswith("/"):
        return ("unix", spec)
    else:
        if spec.startswith("udp:"):
            return ("udp", spec[4:])
        else:
            return ("tcp", spec)

def _unpack_addr(spec, port=None, weight=1):
    if spec.startswith("["):
        if "]" not in spec:
            raise ValueError(spec)
        if spec.endswith("]"):
            addr = spec[1:-1]
        else:
            (addr, port) = spec[1:].rsplit("]:", 1)
    elif ":" in spec:
        (addr, port) = spec.split(":", 1)
    else:
        addr = spec

    if isinstance(port, str) and ":" in port:
        (port, weight) = port.split(":", 1)
    return (addr, port, weight)

def translate_server_spec(server, port=11211, weight=1):
    """Translate/normalize specification *server* into 4-tuple (scheme, addr, port, weight).

    This is format is used by the extension module.

    >>> translate_server_spec("127.0.0.1")
    (1, '127.0.0.1', 11211, 1)
    >>> translate_server_spec("udp:127.0.0.1")
    (2, '127.0.0.1', 11211, 1)
    >>> translate_server_spec("/var/run/memcached.sock")
    (4, '/var/run/memcached.sock', 0, 1)
    >>> translate_server_spec("/var/run/memcached.sock", weight=2)
    (4, '/var/run/memcached.sock', 0, 2)

    >>> translate_server_spec("127.0.0.1:22122")
    (1, '127.0.0.1', 22122, 1)
    >>> translate_server_spec("127.0.0.1:1234:2")
    (1, '127.0.0.1', 1234, 2)
    >>> translate_server_spec("127.0.0.1", port=1234, weight=2)
    (1, '127.0.0.1', 1234, 2)

    >>> translate_server_spec("[::1]:22122")
    (1, '::1', 22122, 1)
    >>> translate_server_spec("[::1]")
    (1, '::1', 11211, 1)
    >>> translate_server_spec("[::1]:22122:3")
    (1, '::1', 22122, 3)
    """
    (scheme, spec) = _split_spec_type(server)
    stype = server_type_map[scheme]
    if scheme == "unix":
        (addr, port) = (spec, 0)
    elif scheme in ("tcp", "udp"):
        (addr, port, weight) = _unpack_addr(spec, port=port, weight=weight)
    return (stype, str(addr), int(port), int(weight))

def translate_server_specs(servers):
    addr_tups = []
    for server in servers:
        # Anti-pattern for convenience
        if isinstance(server, tuple) and len(server) == 3:
            addr_tup = server
        else:
            addr_tup = translate_server_spec(server)
        addr_tups.append(addr_tup)
    return addr_tups

def _behaviors_symbolic(behaviors):
    """Turn numeric constants into symbolic strings"""

    behaviors = behaviors.copy()

    behaviors["hash"] = hashers_rvs[behaviors["hash"]]
    behaviors["distribution"] = distributions_rvs[behaviors["distribution"]]

    return behaviors

def _behaviors_numeric(behaviors):
    """Inverse function of behaviors_symbolic"""

    if not behaviors:
        return behaviors

    behaviors = behaviors.copy()

    if "_retry_timeout" in behaviors:
        behaviors.setdefault("retry_timeout", behaviors.pop("_retry_timeout"))

    unknown = set(behaviors).difference(_all_behaviors_set)
    if unknown:
        names = ", ".join(map(str, sorted(unknown)))
        raise ValueError("unknown behavior names: %s" % (names,))

    if behaviors.get("hash") is not None:
        behaviors["hash"] = hashers[behaviors["hash"]]
    if behaviors.get("ketama_hash") in hashers:
        behaviors["ketama_hash"] = hashers[behaviors["ketama_hash"]]
    if behaviors.get("distribution") is not None:
        behaviors["distribution"] = distributions[behaviors["distribution"]]

    return behaviors

class Client(_pylibmc.client):
    def __init__(self, servers, behaviors=None, binary=False,
                 username=None, password=None):
        """Initialize a memcached client instance.

        This connects to the servers in *servers*, which will default to being
        TCP servers. If it looks like a filesystem path, a UNIX socket. If
        prefixed with `udp:`, a UDP connection.

        If *binary* is True, the binary memcached protocol is used.

        SASL authentication is supported if libmemcached supports it (check
        *pylibmc.support_sasl*). Requires both username and password.
        Note that SASL requires *binary*=True.
        """
        self.binary = binary
        self.addresses = list(servers)
        super(Client, self).__init__(servers=translate_server_specs(servers),
                                     binary=binary,
                                     username=username, password=password,
                                     behaviors=_behaviors_numeric(behaviors))

    def __repr__(self):
        return "%s(%r, binary=%r)" % (self.__class__.__name__,
                                      self.addresses, self.binary)

    def __str__(self):
        addrs = ", ".join(map(str, self.addresses))
        return "<%s for %s, binary=%r>" % (self.__class__.__name__,
                                           addrs, self.binary)

    # {{{ Mapping interface
    def __getitem__(self, key):
        value = self.get(key)
        if value is None:
            raise KeyError(key)
        else:
            return value

    def __setitem__(self, key, value):
        if not self.set(key, value):
            raise KeyError("failed setting %r" % (key,))

    def __delitem__(self, key):
        if not self.delete(key):
            raise KeyError(key)

    def __contains__(self, key):
        return self.get(key) is not None
    # }}}

    # {{{ Behaviors
    def get_behaviors(self):
        """Gets the behaviors from the underlying C client instance.

        Reverses the integer constants for `hash` and `distribution` into more
        understandable string values. See *set_behaviors* for info.
        """
        return BehaviorDict(self, _behaviors_symbolic(super(Client, self).get_behaviors()))

    def set_behaviors(self, behaviors):
        """Sets the behaviors on the underlying C client instance.

        Takes care of morphing the `hash` key, if specified, into the
        corresponding integer constant (which the C client expects.) If,
        however, an unknown value is specified, it's passed on to the C client
        (where it most surely will error out.)

        This also happens for `distribution`.

        Translates old underscored behavior names to new ones for API leniency.
        """
        return super(Client, self).set_behaviors(_behaviors_numeric(behaviors))

    behaviors = property(get_behaviors, set_behaviors)
    @property
    def behaviours(self):
        raise AttributeError("nobody uses british spellings")
    # }}}

    def clone(self):
        obj = super(Client, self).clone()
        obj.addresses = list(self.addresses)
        obj.binary = self.binary
        return obj

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Ported cmemcached tests"""

# NOTE Don't write new tests here.
# These are ported from cmemcached to ensure compatibility.

import pylibmc
from nose.tools import eq_
from tests import PylibmcTestCase

a           = "a"
I_          = b"I "
Do          = b"Do"
I_Do        = b"I Do"
n12345      = "12345"
num12345    = "num12345"
str12345    = "str12345"
hello_world = "hello world"


class TestCmemcached(PylibmcTestCase):
    def testSetAndGet(self):
        self.mc.set(num12345, 12345)
        eq_(self.mc.get(num12345), 12345)
        self.mc.set(str12345, n12345)
        eq_(self.mc.get(str12345), n12345)

    def testDelete(self):
        self.mc.set(str12345, n12345)
        #delete return True on success, otherwise False
        assert self.mc.delete(str12345)
        assert self.mc.get(str12345) is None

        # This test only works with old memcacheds. This has become a "client
        # error" in memcached.
        try:
            assert not self.mc.delete(hello_world)
        except pylibmc.ClientError:
            pass

    def testGetMulti(self):
        self.mc.set("a", "valueA")
        self.mc.set("b", "valueB")
        self.mc.set("c", "valueC")
        result = self.mc.get_multi(["a", "b", "c", "", "hello world"])
        eq_(result, {'a': 'valueA', 'b': 'valueB', 'c': 'valueC'})

    def testBigGetMulti(self):
        count = 10 ** 4
        # Python 2: .encode() is a no-op on these byte strings since they
        # only contain bytes that can be implicitly decoded as ASCII.
        keys = ['key%d' % i for i in range(count)]
        pairs = zip(keys, ['value%d' % i for i in range(count)])

        d = {}
        for key, value in pairs:
            d[key] = value
            self.mc.set(key, value)
        result = self.mc.get_multi(keys)
        eq_(result, d)

    def testFunnyDelete(self):
        s = ""
        assert not self.mc.delete(s)

    def testAppend(self):
        self.mc.delete(a)
        self.mc.set(a, I_)
        assert self.mc.append(a, Do)
        eq_(self.mc.get(a), I_Do)

    def testPrepend(self):
        self.mc.delete(a)
        self.mc.set(a, Do)
        assert self.mc.prepend(a, I_)
        eq_(self.mc.get(a), I_Do)



"""Test utils."""

import os
import socket

import _pylibmc, pylibmc
from textwrap import dedent

class NotAliveError(Exception):
    template = dedent("""
    test memcached %s:%s not alive

    consider setting MEMCACHED_HOST and/or MEMCACHED_PORT
    to something more appropriate.""").lstrip()

    def __init__(self, svr):
        (self.server_type, self.address, self.port) = svr

    def __str__(self):
        return self.template % (self.address, self.port)

def get_version(addr):
    (type_, host, port) = addr
    if type_ != _pylibmc.server_type_tcp:
        raise NotImplementedError("test server can only be on tcp for now")
    else:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((host, port))
            s.send(b"version\r\n")
            version = s.recv(4096)
        finally:
            s.close()
        vstr = b"VERSION "
        rnstr = b"\r\n"
        if not version.startswith(vstr) or not version.endswith(rnstr):
            raise ValueError("unexpected version return: %r" % (version,))
        else:
            version = version[8:-2]
        return version

def is_alive(addr):
    try:
        version = get_version(addr)
    except (ValueError, socket.error):
        version = None
    return bool(version), version

def make_test_client(cls=pylibmc.Client, env=None, server_type="tcp",
                     host="127.0.0.1", port=11211, **kwds):
    """Make a test client. `env` overrides keyword arguments.

    >>> make_test_client() is not None
    True
    """
    server_types = {"tcp": _pylibmc.server_type_tcp,
                    "udp": _pylibmc.server_type_udp,
                    "unix": _pylibmc.server_type_unix}

    if env is None:
        env = os.environ

    styp = env.get("MEMCACHED_TYPE", server_type)
    host = env.get("MEMCACHED_HOST", host)
    port = env.get("MEMCACHED_PORT", port)

    test_server = (server_types[styp], str(host), int(port))
    alive, version = is_alive(test_server)

    if alive:
        return cls([test_server], **kwds)
    else:
        raise NotAliveError(test_server)



import functools
import sys
import time
import pylibmc
import _pylibmc
from pylibmc.test import make_test_client
from tests import PylibmcTestCase
from nose import SkipTest
from nose.tools import eq_, ok_

PY3 = sys.version_info[0] >= 3

def requires_memcached_touch(test):
    @functools.wraps(test)
    def wrapper(*args, **kwargs):
        if _pylibmc.libmemcached_version_hex >= 0x01000002:
            return test(*args, **kwargs)
        raise SkipTest

    return wrapper

class ClientTests(PylibmcTestCase):
    def test_zerokey(self):
        bc = make_test_client(binary=True)
        k = "\x00\x01"
        test_str = "test"
        ok_(bc.set(k, test_str))
        rk = next(iter(bc.get_multi([k])))
        eq_(k, rk)

    def test_cas(self):
        c = "cas"
        k = "testkey"
        mc = make_test_client(binary=False, behaviors={c: True})
        ok_(mc.set(k, 0))
        while True:
            rv, cas = mc.gets(k)
            ok_(mc.cas(k, rv + 1, cas))
            if rv == 10:
                break

    def testBehaviors(self):
        expected_behaviors = [
            'auto_eject', 'buffer_requests', 'cas', 'connect_timeout',
            'distribution', 'failure_limit', 'hash', 'ketama', 'ketama_hash',
            'ketama_weighted', 'no_block', 'num_replicas', 'receive_timeout',
            'retry_timeout', 'send_timeout', 'tcp_keepalive', 'tcp_nodelay',
            'verify_keys']

        # Since some parts of pyblibmc's functionality depend on the
        # libmemcached version, programatically check for the expected values
        # (see _pylibmcmodule.h for the complete list of #ifdef versions)
        if _pylibmc.libmemcached_version_hex >= 0x00049000:
            expected_behaviors.append("remove_failed")
        if _pylibmc.libmemcached_version_hex >= 0x01000003:
            expected_behaviors.append("dead_timeout")

        # Filter out private keys
        actual_behaviors = [
                behavior for behavior in self.mc.behaviors
                if not behavior.startswith('_')]

        sorted_list = lambda L: list(sorted(L))
        eq_(sorted_list(expected_behaviors),
            sorted_list(actual_behaviors))

    @requires_memcached_touch
    def test_touch(self):
        touch_test  = "touch-test"
        touch_test2 = "touch-test-2"
        tval = "touch-val"
        ok_(self.mc.set(touch_test, tval, 1))
        eq_(tval, self.mc.get(touch_test))
        time.sleep(2)
        eq_(None, self.mc.get(touch_test))

        self.mc.set(touch_test, tval, 1)
        eq_(tval, self.mc.get(touch_test))
        ok_(self.mc.touch(touch_test, 5))
        time.sleep(2)
        eq_(tval, self.mc.get(touch_test))

        ok_(not self.mc.touch(touch_test2, 100))

    def test_exceptions(self):
        self.assertRaises(TypeError, self.mc.set, 1, "hi")
        self.assertRaises(_pylibmc.Error, _pylibmc.client, [])
        self.assertRaises(_pylibmc.NotFound, self.mc.incr_multi,
            ('a', 'b', 'c'), key_prefix='x', delta=1)

    def test_utf8_encoding(self):
        k = "a key with a replacement character \ufffd and something non-BMP \U0001f4a3"
        k_enc = k.encode('utf-8')
        mc = make_test_client(binary=True)
        ok_(mc.set(k, 0))
        ok_(mc.get(k_enc) == 0)



from __future__ import print_function
import os
import sys
from distutils.core import setup, Extension

# Need an 'open' function that supports the 'encoding' argument:
if sys.version_info[0] < 3:
    from codecs import open

## Command-line argument parsing

# --with-zlib: use zlib for compressing and decompressing
# --without-zlib: ^ negated
# --with-zlib=<dir>: path to zlib if needed
# --with-libmemcached=<dir>: path to libmemcached package if needed

cmd = None
use_zlib = True
pkgdirs = []  # incdirs and libdirs get these
libs = ["memcached"]
defs = []
incdirs = []
libdirs = []

def append_env(L, e):
    v = os.environ.get(e)
    if v and os.path.exists(v):
        L.append(v)

append_env(pkgdirs, "LIBMEMCACHED")
append_env(pkgdirs, "ZLIB")

# Hack up sys.argv, yay
unprocessed = []
for arg in sys.argv[1:]:
    if arg == "--with-zlib":
        use_zlib = True
        continue
    elif arg == "--without-zlib":
        use_zlib = False
        continue
    elif arg == "--with-sasl2":
        libs.append("sasl2")
        continue
    elif arg == "--gen-setup":
        cmd = arg[2:]
    elif "=" in arg:
        if arg.startswith("--with-libmemcached=") or \
           arg.startswith("--with-zlib="):
            pkgdirs.append(arg.split("=", 1)[1])
            continue
    unprocessed.append(arg)
sys.argv[1:] = unprocessed

for pkgdir in pkgdirs:
    incdirs.append(os.path.join(pkgdir, "include"))
    libdirs.append(os.path.join(pkgdir, "lib"))

if use_zlib:
    libs.append("z")
    defs.append(("USE_ZLIB", None))

## OS X non-PPC workaround

# Apple OS X 10.6 with Xcode 4 have Python compiled with PPC but they removed
# support for compiling with that arch, so we have to override ARCHFLAGS.
if sys.platform == "darwin" and not os.environ.get("ARCHFLAGS"):
    compiler_dirn = "/usr/libexec/gcc/darwin"
    if os.path.exists(compiler_dirn):
        dir_items = os.listdir(compiler_dirn)
        if "ppc" not in dir_items:
            print("enabling osx-specific ARCHFLAGS/ppc hack", file=sys.stderr)
            os.environ["ARCHFLAGS"] = "-arch i386 -arch x86_64"

# There's a bug in <py3 with Py_True/False that will propagate with GCC's
# strict aliasing rules. Let's skip this flag for now.
cflags = ["-fno-strict-aliasing", ]

## Extension definitions

pylibmc_ext = Extension("_pylibmc", ["src/_pylibmcmodule.c"],
                        libraries=libs, include_dirs=incdirs,
                        library_dirs=libdirs, define_macros=defs,
                        extra_compile_args=cflags)

# Hidden secret: if environment variable GEN_SETUP is set, generate Setup file.
if cmd == "gen-setup":
    line = " ".join((
        pylibmc_ext.name,
        " ".join("-l" + lib for lib in pylibmc_ext.libraries),
        " ".join("-I" + incdir for incdir in pylibmc_ext.include_dirs),
        " ".join("-L" + libdir for libdir in pylibmc_ext.library_dirs),
        " ".join("-D" + name + ("=" + str(value), "")[value is None] for (name, value) in pylibmc_ext.define_macros)))
    with open("Setup", "w") as s:
        s.write(line + "\n")
    sys.exit(0)

with open("README.rst", "U", encoding="utf-8") as r:
    readme_text = r.read()
with open("src/pylibmc-version.h", "U", encoding="utf-8") as r:
    version = r.read().strip().split("\"")[1]

setup(
    name="pylibmc",
    version=version,
    url="http://sendapatch.se/projects/pylibmc/",
    author="Ludvig Ericson",
    author_email="ludvig@lericson.se",
    license="3-clause BSD <http://www.opensource.org/licenses/bsd-license.php>",
    description="Quick and small memcached client for Python",
    long_description=readme_text,
    ext_modules=[pylibmc_ext],
    package_dir={'': 'src'},
    packages=['pylibmc'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)
